/* Generated with `python conf/mpiapigen.py` */
#ifndef PyMPI_PYMPICONF_H
#define PyMPI_PYMPICONF_H

#undef  PyMPI_HAVE_MPI_Aint
#undef  PyMPI_HAVE_MPI_Offset
#undef  PyMPI_HAVE_MPI_Count
#undef  PyMPI_HAVE_MPI_Status
#undef  PyMPI_HAVE_MPI_Datatype
#undef  PyMPI_HAVE_MPI_Request
#undef  PyMPI_HAVE_MPI_Message
#undef  PyMPI_HAVE_MPI_Op
#undef  PyMPI_HAVE_MPI_Group
#undef  PyMPI_HAVE_MPI_Info
#undef  PyMPI_HAVE_MPI_Errhandler
#undef  PyMPI_HAVE_MPI_Session
#undef  PyMPI_HAVE_MPI_Comm
#undef  PyMPI_HAVE_MPI_Win
#undef  PyMPI_HAVE_MPI_File
#undef  PyMPI_HAVE_MPI_UNDEFINED
#undef  PyMPI_HAVE_MPI_ANY_SOURCE
#undef  PyMPI_HAVE_MPI_ANY_TAG
#undef  PyMPI_HAVE_MPI_PROC_NULL
#undef  PyMPI_HAVE_MPI_ROOT
#undef  PyMPI_HAVE_MPI_IDENT
#undef  PyMPI_HAVE_MPI_CONGRUENT
#undef  PyMPI_HAVE_MPI_SIMILAR
#undef  PyMPI_HAVE_MPI_UNEQUAL
#undef  PyMPI_HAVE_MPI_BOTTOM
#undef  PyMPI_HAVE_MPI_IN_PLACE
#undef  PyMPI_HAVE_MPI_KEYVAL_INVALID
#undef  PyMPI_HAVE_MPI_MAX_OBJECT_NAME
#undef  PyMPI_HAVE_MPI_DATATYPE_NULL
#undef  PyMPI_HAVE_MPI_PACKED
#undef  PyMPI_HAVE_MPI_BYTE
#undef  PyMPI_HAVE_MPI_AINT
#undef  PyMPI_HAVE_MPI_OFFSET
#undef  PyMPI_HAVE_MPI_COUNT
#undef  PyMPI_HAVE_MPI_CHAR
#undef  PyMPI_HAVE_MPI_WCHAR
#undef  PyMPI_HAVE_MPI_SIGNED_CHAR
#undef  PyMPI_HAVE_MPI_SHORT
#undef  PyMPI_HAVE_MPI_INT
#undef  PyMPI_HAVE_MPI_LONG
#undef  PyMPI_HAVE_MPI_LONG_LONG
#undef  PyMPI_HAVE_MPI_LONG_LONG_INT
#undef  PyMPI_HAVE_MPI_UNSIGNED_CHAR
#undef  PyMPI_HAVE_MPI_UNSIGNED_SHORT
#undef  PyMPI_HAVE_MPI_UNSIGNED
#undef  PyMPI_HAVE_MPI_UNSIGNED_LONG
#undef  PyMPI_HAVE_MPI_UNSIGNED_LONG_LONG
#undef  PyMPI_HAVE_MPI_FLOAT
#undef  PyMPI_HAVE_MPI_DOUBLE
#undef  PyMPI_HAVE_MPI_LONG_DOUBLE
#undef  PyMPI_HAVE_MPI_C_BOOL
#undef  PyMPI_HAVE_MPI_INT8_T
#undef  PyMPI_HAVE_MPI_INT16_T
#undef  PyMPI_HAVE_MPI_INT32_T
#undef  PyMPI_HAVE_MPI_INT64_T
#undef  PyMPI_HAVE_MPI_UINT8_T
#undef  PyMPI_HAVE_MPI_UINT16_T
#undef  PyMPI_HAVE_MPI_UINT32_T
#undef  PyMPI_HAVE_MPI_UINT64_T
#undef  PyMPI_HAVE_MPI_FLOAT16_T
#undef  PyMPI_HAVE_MPI_BFLOAT16_T
#undef  PyMPI_HAVE_MPI_C_COMPLEX
#undef  PyMPI_HAVE_MPI_C_FLOAT_COMPLEX
#undef  PyMPI_HAVE_MPI_C_DOUBLE_COMPLEX
#undef  PyMPI_HAVE_MPI_C_LONG_DOUBLE_COMPLEX
#undef  PyMPI_HAVE_MPI_CXX_BOOL
#undef  PyMPI_HAVE_MPI_CXX_FLOAT_COMPLEX
#undef  PyMPI_HAVE_MPI_CXX_DOUBLE_COMPLEX
#undef  PyMPI_HAVE_MPI_CXX_LONG_DOUBLE_COMPLEX
#undef  PyMPI_HAVE_MPI_SHORT_INT
#undef  PyMPI_HAVE_MPI_2INT
#undef  PyMPI_HAVE_MPI_LONG_INT
#undef  PyMPI_HAVE_MPI_FLOAT_INT
#undef  PyMPI_HAVE_MPI_DOUBLE_INT
#undef  PyMPI_HAVE_MPI_LONG_DOUBLE_INT
#undef  PyMPI_HAVE_MPI_CHARACTER
#undef  PyMPI_HAVE_MPI_LOGICAL
#undef  PyMPI_HAVE_MPI_INTEGER
#undef  PyMPI_HAVE_MPI_REAL
#undef  PyMPI_HAVE_MPI_DOUBLE_PRECISION
#undef  PyMPI_HAVE_MPI_COMPLEX
#undef  PyMPI_HAVE_MPI_DOUBLE_COMPLEX
#undef  PyMPI_HAVE_MPI_LOGICAL1
#undef  PyMPI_HAVE_MPI_LOGICAL2
#undef  PyMPI_HAVE_MPI_LOGICAL4
#undef  PyMPI_HAVE_MPI_LOGICAL8
#undef  PyMPI_HAVE_MPI_LOGICAL16
#undef  PyMPI_HAVE_MPI_INTEGER1
#undef  PyMPI_HAVE_MPI_INTEGER2
#undef  PyMPI_HAVE_MPI_INTEGER4
#undef  PyMPI_HAVE_MPI_INTEGER8
#undef  PyMPI_HAVE_MPI_INTEGER16
#undef  PyMPI_HAVE_MPI_REAL2
#undef  PyMPI_HAVE_MPI_REAL4
#undef  PyMPI_HAVE_MPI_REAL8
#undef  PyMPI_HAVE_MPI_REAL16
#undef  PyMPI_HAVE_MPI_COMPLEX4
#undef  PyMPI_HAVE_MPI_COMPLEX8
#undef  PyMPI_HAVE_MPI_COMPLEX16
#undef  PyMPI_HAVE_MPI_COMPLEX32
#undef  PyMPI_HAVE_MPI_Get_address
#undef  PyMPI_HAVE_MPI_Aint_add
#undef  PyMPI_HAVE_MPI_Aint_diff
#undef  PyMPI_HAVE_MPI_Type_dup
#undef  PyMPI_HAVE_MPI_Type_contiguous
#undef  PyMPI_HAVE_MPI_Type_vector
#undef  PyMPI_HAVE_MPI_Type_indexed
#undef  PyMPI_HAVE_MPI_Type_create_indexed_block
#undef  PyMPI_HAVE_MPI_ORDER_C
#undef  PyMPI_HAVE_MPI_ORDER_FORTRAN
#undef  PyMPI_HAVE_MPI_Type_create_subarray
#undef  PyMPI_HAVE_MPI_DISTRIBUTE_NONE
#undef  PyMPI_HAVE_MPI_DISTRIBUTE_BLOCK
#undef  PyMPI_HAVE_MPI_DISTRIBUTE_CYCLIC
#undef  PyMPI_HAVE_MPI_DISTRIBUTE_DFLT_DARG
#undef  PyMPI_HAVE_MPI_Type_create_darray
#undef  PyMPI_HAVE_MPI_Type_create_hvector
#undef  PyMPI_HAVE_MPI_Type_create_hindexed
#undef  PyMPI_HAVE_MPI_Type_create_hindexed_block
#undef  PyMPI_HAVE_MPI_Type_create_struct
#undef  PyMPI_HAVE_MPI_Type_create_resized
#undef  PyMPI_HAVE_MPI_Type_size
#undef  PyMPI_HAVE_MPI_Type_get_extent
#undef  PyMPI_HAVE_MPI_Type_get_true_extent
#undef  PyMPI_HAVE_MPI_Type_size_x
#undef  PyMPI_HAVE_MPI_Type_get_extent_x
#undef  PyMPI_HAVE_MPI_Type_get_true_extent_x
#undef  PyMPI_HAVE_MPI_Type_create_f90_integer
#undef  PyMPI_HAVE_MPI_Type_create_f90_real
#undef  PyMPI_HAVE_MPI_Type_create_f90_complex
#undef  PyMPI_HAVE_MPI_TYPECLASS_LOGICAL
#undef  PyMPI_HAVE_MPI_TYPECLASS_INTEGER
#undef  PyMPI_HAVE_MPI_TYPECLASS_REAL
#undef  PyMPI_HAVE_MPI_TYPECLASS_COMPLEX
#undef  PyMPI_HAVE_MPI_Type_match_size
#undef  PyMPI_HAVE_MPI_Type_get_value_index
#undef  PyMPI_HAVE_MPI_Type_commit
#undef  PyMPI_HAVE_MPI_Type_free
#undef  PyMPI_HAVE_MPI_COMBINER_NAMED
#undef  PyMPI_HAVE_MPI_COMBINER_DUP
#undef  PyMPI_HAVE_MPI_COMBINER_CONTIGUOUS
#undef  PyMPI_HAVE_MPI_COMBINER_VECTOR
#undef  PyMPI_HAVE_MPI_COMBINER_HVECTOR
#undef  PyMPI_HAVE_MPI_COMBINER_INDEXED
#undef  PyMPI_HAVE_MPI_COMBINER_HINDEXED
#undef  PyMPI_HAVE_MPI_COMBINER_INDEXED_BLOCK
#undef  PyMPI_HAVE_MPI_COMBINER_HINDEXED_BLOCK
#undef  PyMPI_HAVE_MPI_COMBINER_STRUCT
#undef  PyMPI_HAVE_MPI_COMBINER_SUBARRAY
#undef  PyMPI_HAVE_MPI_COMBINER_DARRAY
#undef  PyMPI_HAVE_MPI_COMBINER_F90_REAL
#undef  PyMPI_HAVE_MPI_COMBINER_F90_COMPLEX
#undef  PyMPI_HAVE_MPI_COMBINER_F90_INTEGER
#undef  PyMPI_HAVE_MPI_COMBINER_RESIZED
#undef  PyMPI_HAVE_MPI_COMBINER_VALUE_INDEX
#undef  PyMPI_HAVE_MPI_Type_get_envelope
#undef  PyMPI_HAVE_MPI_Type_get_contents
#undef  PyMPI_HAVE_MPI_Pack
#undef  PyMPI_HAVE_MPI_Unpack
#undef  PyMPI_HAVE_MPI_Pack_size
#undef  PyMPI_HAVE_MPI_Pack_external
#undef  PyMPI_HAVE_MPI_Unpack_external
#undef  PyMPI_HAVE_MPI_Pack_external_size
#undef  PyMPI_HAVE_MPI_Type_get_name
#undef  PyMPI_HAVE_MPI_Type_set_name
#undef  PyMPI_HAVE_MPI_Type_get_attr
#undef  PyMPI_HAVE_MPI_Type_set_attr
#undef  PyMPI_HAVE_MPI_Type_delete_attr
#undef  PyMPI_HAVE_MPI_Type_copy_attr_function
#undef  PyMPI_HAVE_MPI_Type_delete_attr_function
#undef  PyMPI_HAVE_MPI_TYPE_NULL_COPY_FN
#undef  PyMPI_HAVE_MPI_TYPE_DUP_FN
#undef  PyMPI_HAVE_MPI_TYPE_NULL_DELETE_FN
#undef  PyMPI_HAVE_MPI_Type_create_keyval
#undef  PyMPI_HAVE_MPI_Type_free_keyval
#undef  PyMPI_HAVE_MPI_Type_contiguous_c
#undef  PyMPI_HAVE_MPI_Type_vector_c
#undef  PyMPI_HAVE_MPI_Type_indexed_c
#undef  PyMPI_HAVE_MPI_Type_create_indexed_block_c
#undef  PyMPI_HAVE_MPI_Type_create_subarray_c
#undef  PyMPI_HAVE_MPI_Type_create_darray_c
#undef  PyMPI_HAVE_MPI_Type_create_hvector_c
#undef  PyMPI_HAVE_MPI_Type_create_hindexed_c
#undef  PyMPI_HAVE_MPI_Type_create_hindexed_block_c
#undef  PyMPI_HAVE_MPI_Type_create_struct_c
#undef  PyMPI_HAVE_MPI_Type_create_resized_c
#undef  PyMPI_HAVE_MPI_Type_size_c
#undef  PyMPI_HAVE_MPI_Type_get_extent_c
#undef  PyMPI_HAVE_MPI_Type_get_true_extent_c
#undef  PyMPI_HAVE_MPI_Type_get_envelope_c
#undef  PyMPI_HAVE_MPI_Type_get_contents_c
#undef  PyMPI_HAVE_MPI_Pack_c
#undef  PyMPI_HAVE_MPI_Unpack_c
#undef  PyMPI_HAVE_MPI_Pack_size_c
#undef  PyMPI_HAVE_MPI_Pack_external_c
#undef  PyMPI_HAVE_MPI_Unpack_external_c
#undef  PyMPI_HAVE_MPI_Pack_external_size_c
#undef  PyMPI_HAVE_MPI_STATUS_IGNORE
#undef  PyMPI_HAVE_MPI_STATUSES_IGNORE
#undef  PyMPI_HAVE_MPI_Get_count
#undef  PyMPI_HAVE_MPI_Get_elements
#undef  PyMPI_HAVE_MPI_Status_set_elements
#undef  PyMPI_HAVE_MPI_Get_elements_x
#undef  PyMPI_HAVE_MPI_Status_set_elements_x
#undef  PyMPI_HAVE_MPI_Test_cancelled
#undef  PyMPI_HAVE_MPI_Status_set_cancelled
#undef  PyMPI_HAVE_MPI_Get_count_c
#undef  PyMPI_HAVE_MPI_Get_elements_c
#undef  PyMPI_HAVE_MPI_Status_set_elements_c
#undef  PyMPI_HAVE_MPI_Status_get_source
#undef  PyMPI_HAVE_MPI_Status_set_source
#undef  PyMPI_HAVE_MPI_Status_get_tag
#undef  PyMPI_HAVE_MPI_Status_set_tag
#undef  PyMPI_HAVE_MPI_Status_get_error
#undef  PyMPI_HAVE_MPI_Status_set_error
#undef  PyMPI_HAVE_MPI_REQUEST_NULL
#undef  PyMPI_HAVE_MPI_Wait
#undef  PyMPI_HAVE_MPI_Test
#undef  PyMPI_HAVE_MPI_Request_get_status
#undef  PyMPI_HAVE_MPI_Waitany
#undef  PyMPI_HAVE_MPI_Testany
#undef  PyMPI_HAVE_MPI_Request_get_status_any
#undef  PyMPI_HAVE_MPI_Waitall
#undef  PyMPI_HAVE_MPI_Testall
#undef  PyMPI_HAVE_MPI_Request_get_status_all
#undef  PyMPI_HAVE_MPI_Waitsome
#undef  PyMPI_HAVE_MPI_Testsome
#undef  PyMPI_HAVE_MPI_Request_get_status_some
#undef  PyMPI_HAVE_MPI_Cancel
#undef  PyMPI_HAVE_MPI_Request_free
#undef  PyMPI_HAVE_MPI_Start
#undef  PyMPI_HAVE_MPI_Startall
#undef  PyMPI_HAVE_MPI_Pready
#undef  PyMPI_HAVE_MPI_Pready_range
#undef  PyMPI_HAVE_MPI_Pready_list
#undef  PyMPI_HAVE_MPI_Parrived
#undef  PyMPI_HAVE_MPI_Grequest_cancel_function
#undef  PyMPI_HAVE_MPI_Grequest_free_function
#undef  PyMPI_HAVE_MPI_Grequest_query_function
#undef  PyMPI_HAVE_MPI_Grequest_start
#undef  PyMPI_HAVE_MPI_Grequest_complete
#undef  PyMPI_HAVE_MPI_OP_NULL
#undef  PyMPI_HAVE_MPI_MAX
#undef  PyMPI_HAVE_MPI_MIN
#undef  PyMPI_HAVE_MPI_SUM
#undef  PyMPI_HAVE_MPI_PROD
#undef  PyMPI_HAVE_MPI_LAND
#undef  PyMPI_HAVE_MPI_BAND
#undef  PyMPI_HAVE_MPI_LOR
#undef  PyMPI_HAVE_MPI_BOR
#undef  PyMPI_HAVE_MPI_LXOR
#undef  PyMPI_HAVE_MPI_BXOR
#undef  PyMPI_HAVE_MPI_MAXLOC
#undef  PyMPI_HAVE_MPI_MINLOC
#undef  PyMPI_HAVE_MPI_REPLACE
#undef  PyMPI_HAVE_MPI_NO_OP
#undef  PyMPI_HAVE_MPI_Op_free
#undef  PyMPI_HAVE_MPI_User_function
#undef  PyMPI_HAVE_MPI_Op_create
#undef  PyMPI_HAVE_MPI_Op_commutative
#undef  PyMPI_HAVE_MPI_User_function_c
#undef  PyMPI_HAVE_MPI_Op_create_c
#undef  PyMPI_HAVE_MPI_GROUP_NULL
#undef  PyMPI_HAVE_MPI_GROUP_EMPTY
#undef  PyMPI_HAVE_MPI_Group_free
#undef  PyMPI_HAVE_MPI_Group_size
#undef  PyMPI_HAVE_MPI_Group_rank
#undef  PyMPI_HAVE_MPI_Group_translate_ranks
#undef  PyMPI_HAVE_MPI_Group_compare
#undef  PyMPI_HAVE_MPI_Group_union
#undef  PyMPI_HAVE_MPI_Group_intersection
#undef  PyMPI_HAVE_MPI_Group_difference
#undef  PyMPI_HAVE_MPI_Group_incl
#undef  PyMPI_HAVE_MPI_Group_excl
#undef  PyMPI_HAVE_MPI_Group_range_incl
#undef  PyMPI_HAVE_MPI_Group_range_excl
#undef  PyMPI_HAVE_MPI_INFO_NULL
#undef  PyMPI_HAVE_MPI_INFO_ENV
#undef  PyMPI_HAVE_MPI_Info_free
#undef  PyMPI_HAVE_MPI_Info_create
#undef  PyMPI_HAVE_MPI_Info_dup
#undef  PyMPI_HAVE_MPI_Info_create_env
#undef  PyMPI_HAVE_MPI_MAX_INFO_KEY
#undef  PyMPI_HAVE_MPI_MAX_INFO_VAL
#undef  PyMPI_HAVE_MPI_Info_get_string
#undef  PyMPI_HAVE_MPI_Info_set
#undef  PyMPI_HAVE_MPI_Info_delete
#undef  PyMPI_HAVE_MPI_Info_get_nkeys
#undef  PyMPI_HAVE_MPI_Info_get_nthkey
#undef  PyMPI_HAVE_MPI_ERRHANDLER_NULL
#undef  PyMPI_HAVE_MPI_ERRORS_RETURN
#undef  PyMPI_HAVE_MPI_ERRORS_ABORT
#undef  PyMPI_HAVE_MPI_ERRORS_ARE_FATAL
#undef  PyMPI_HAVE_MPI_Errhandler_free
#undef  PyMPI_HAVE_MPI_SESSION_NULL
#undef  PyMPI_HAVE_MPI_MAX_PSET_NAME_LEN
#undef  PyMPI_HAVE_MPI_Session_init
#undef  PyMPI_HAVE_MPI_Session_finalize
#undef  PyMPI_HAVE_MPI_Session_get_num_psets
#undef  PyMPI_HAVE_MPI_Session_get_nth_pset
#undef  PyMPI_HAVE_MPI_Session_get_info
#undef  PyMPI_HAVE_MPI_Session_get_pset_info
#undef  PyMPI_HAVE_MPI_Group_from_session_pset
#undef  PyMPI_HAVE_MPI_Session_errhandler_function
#undef  PyMPI_HAVE_MPI_Session_create_errhandler
#undef  PyMPI_HAVE_MPI_Session_get_errhandler
#undef  PyMPI_HAVE_MPI_Session_set_errhandler
#undef  PyMPI_HAVE_MPI_Session_call_errhandler
#undef  PyMPI_HAVE_MPI_COMM_NULL
#undef  PyMPI_HAVE_MPI_COMM_SELF
#define PyMPI_HAVE_MPI_COMM_WORLD 1
#undef  PyMPI_HAVE_MPI_Comm_free
#undef  PyMPI_HAVE_MPI_Comm_group
#define PyMPI_HAVE_MPI_Comm_size 1
#define PyMPI_HAVE_MPI_Comm_rank 1
#undef  PyMPI_HAVE_MPI_Comm_compare
#undef  PyMPI_HAVE_MPI_Topo_test
#undef  PyMPI_HAVE_MPI_Comm_test_inter
#define PyMPI_HAVE_MPI_Abort 1
#undef  PyMPI_HAVE_MPI_BSEND_OVERHEAD
#undef  PyMPI_HAVE_MPI_BUFFER_AUTOMATIC
#undef  PyMPI_HAVE_MPI_Buffer_attach
#undef  PyMPI_HAVE_MPI_Buffer_detach
#undef  PyMPI_HAVE_MPI_Buffer_flush
#undef  PyMPI_HAVE_MPI_Buffer_iflush
#undef  PyMPI_HAVE_MPI_Comm_attach_buffer
#undef  PyMPI_HAVE_MPI_Comm_detach_buffer
#undef  PyMPI_HAVE_MPI_Comm_flush_buffer
#undef  PyMPI_HAVE_MPI_Comm_iflush_buffer
#undef  PyMPI_HAVE_MPI_Session_attach_buffer
#undef  PyMPI_HAVE_MPI_Session_detach_buffer
#undef  PyMPI_HAVE_MPI_Session_flush_buffer
#undef  PyMPI_HAVE_MPI_Session_iflush_buffer
#undef  PyMPI_HAVE_MPI_Send
#undef  PyMPI_HAVE_MPI_Recv
#undef  PyMPI_HAVE_MPI_Sendrecv
#undef  PyMPI_HAVE_MPI_Sendrecv_replace
#undef  PyMPI_HAVE_MPI_Bsend
#undef  PyMPI_HAVE_MPI_Ssend
#undef  PyMPI_HAVE_MPI_Rsend
#undef  PyMPI_HAVE_MPI_Isend
#undef  PyMPI_HAVE_MPI_Irecv
#undef  PyMPI_HAVE_MPI_Isendrecv
#undef  PyMPI_HAVE_MPI_Isendrecv_replace
#undef  PyMPI_HAVE_MPI_Ibsend
#undef  PyMPI_HAVE_MPI_Issend
#undef  PyMPI_HAVE_MPI_Irsend
#undef  PyMPI_HAVE_MPI_Send_init
#undef  PyMPI_HAVE_MPI_Bsend_init
#undef  PyMPI_HAVE_MPI_Ssend_init
#undef  PyMPI_HAVE_MPI_Rsend_init
#undef  PyMPI_HAVE_MPI_Recv_init
#undef  PyMPI_HAVE_MPI_Psend_init
#undef  PyMPI_HAVE_MPI_Precv_init
#undef  PyMPI_HAVE_MPI_Probe
#undef  PyMPI_HAVE_MPI_Iprobe
#undef  PyMPI_HAVE_MPI_MESSAGE_NULL
#undef  PyMPI_HAVE_MPI_MESSAGE_NO_PROC
#undef  PyMPI_HAVE_MPI_Mprobe
#undef  PyMPI_HAVE_MPI_Improbe
#undef  PyMPI_HAVE_MPI_Mrecv
#undef  PyMPI_HAVE_MPI_Imrecv
#undef  PyMPI_HAVE_MPI_Barrier
#undef  PyMPI_HAVE_MPI_Bcast
#undef  PyMPI_HAVE_MPI_Gather
#undef  PyMPI_HAVE_MPI_Gatherv
#undef  PyMPI_HAVE_MPI_Scatter
#undef  PyMPI_HAVE_MPI_Scatterv
#undef  PyMPI_HAVE_MPI_Allgather
#undef  PyMPI_HAVE_MPI_Allgatherv
#undef  PyMPI_HAVE_MPI_Alltoall
#undef  PyMPI_HAVE_MPI_Alltoallv
#undef  PyMPI_HAVE_MPI_Alltoallw
#undef  PyMPI_HAVE_MPI_Reduce_local
#undef  PyMPI_HAVE_MPI_Reduce
#undef  PyMPI_HAVE_MPI_Allreduce
#undef  PyMPI_HAVE_MPI_Reduce_scatter_block
#undef  PyMPI_HAVE_MPI_Reduce_scatter
#undef  PyMPI_HAVE_MPI_Scan
#undef  PyMPI_HAVE_MPI_Exscan
#undef  PyMPI_HAVE_MPI_Neighbor_allgather
#undef  PyMPI_HAVE_MPI_Neighbor_allgatherv
#undef  PyMPI_HAVE_MPI_Neighbor_alltoall
#undef  PyMPI_HAVE_MPI_Neighbor_alltoallv
#undef  PyMPI_HAVE_MPI_Neighbor_alltoallw
#undef  PyMPI_HAVE_MPI_Ibarrier
#undef  PyMPI_HAVE_MPI_Ibcast
#undef  PyMPI_HAVE_MPI_Igather
#undef  PyMPI_HAVE_MPI_Igatherv
#undef  PyMPI_HAVE_MPI_Iscatter
#undef  PyMPI_HAVE_MPI_Iscatterv
#undef  PyMPI_HAVE_MPI_Iallgather
#undef  PyMPI_HAVE_MPI_Iallgatherv
#undef  PyMPI_HAVE_MPI_Ialltoall
#undef  PyMPI_HAVE_MPI_Ialltoallv
#undef  PyMPI_HAVE_MPI_Ialltoallw
#undef  PyMPI_HAVE_MPI_Ireduce
#undef  PyMPI_HAVE_MPI_Iallreduce
#undef  PyMPI_HAVE_MPI_Ireduce_scatter_block
#undef  PyMPI_HAVE_MPI_Ireduce_scatter
#undef  PyMPI_HAVE_MPI_Iscan
#undef  PyMPI_HAVE_MPI_Iexscan
#undef  PyMPI_HAVE_MPI_Ineighbor_allgather
#undef  PyMPI_HAVE_MPI_Ineighbor_allgatherv
#undef  PyMPI_HAVE_MPI_Ineighbor_alltoall
#undef  PyMPI_HAVE_MPI_Ineighbor_alltoallv
#undef  PyMPI_HAVE_MPI_Ineighbor_alltoallw
#undef  PyMPI_HAVE_MPI_Barrier_init
#undef  PyMPI_HAVE_MPI_Bcast_init
#undef  PyMPI_HAVE_MPI_Gather_init
#undef  PyMPI_HAVE_MPI_Gatherv_init
#undef  PyMPI_HAVE_MPI_Scatter_init
#undef  PyMPI_HAVE_MPI_Scatterv_init
#undef  PyMPI_HAVE_MPI_Allgather_init
#undef  PyMPI_HAVE_MPI_Allgatherv_init
#undef  PyMPI_HAVE_MPI_Alltoall_init
#undef  PyMPI_HAVE_MPI_Alltoallv_init
#undef  PyMPI_HAVE_MPI_Alltoallw_init
#undef  PyMPI_HAVE_MPI_Reduce_init
#undef  PyMPI_HAVE_MPI_Allreduce_init
#undef  PyMPI_HAVE_MPI_Reduce_scatter_block_init
#undef  PyMPI_HAVE_MPI_Reduce_scatter_init
#undef  PyMPI_HAVE_MPI_Scan_init
#undef  PyMPI_HAVE_MPI_Exscan_init
#undef  PyMPI_HAVE_MPI_Neighbor_allgather_init
#undef  PyMPI_HAVE_MPI_Neighbor_allgatherv_init
#undef  PyMPI_HAVE_MPI_Neighbor_alltoall_init
#undef  PyMPI_HAVE_MPI_Neighbor_alltoallv_init
#undef  PyMPI_HAVE_MPI_Neighbor_alltoallw_init
#undef  PyMPI_HAVE_MPI_Comm_dup
#undef  PyMPI_HAVE_MPI_Comm_dup_with_info
#undef  PyMPI_HAVE_MPI_Comm_idup
#undef  PyMPI_HAVE_MPI_Comm_idup_with_info
#undef  PyMPI_HAVE_MPI_Comm_create
#undef  PyMPI_HAVE_MPI_Comm_create_group
#undef  PyMPI_HAVE_MPI_MAX_STRINGTAG_LEN
#undef  PyMPI_HAVE_MPI_Comm_create_from_group
#undef  PyMPI_HAVE_MPI_Comm_split
#undef  PyMPI_HAVE_MPI_COMM_TYPE_SHARED
#undef  PyMPI_HAVE_MPI_COMM_TYPE_HW_GUIDED
#undef  PyMPI_HAVE_MPI_COMM_TYPE_HW_UNGUIDED
#undef  PyMPI_HAVE_MPI_COMM_TYPE_RESOURCE_GUIDED
#undef  PyMPI_HAVE_MPI_Comm_split_type
#undef  PyMPI_HAVE_MPI_Comm_set_info
#undef  PyMPI_HAVE_MPI_Comm_get_info
#undef  PyMPI_HAVE_MPI_CART
#undef  PyMPI_HAVE_MPI_Cart_create
#undef  PyMPI_HAVE_MPI_Cartdim_get
#undef  PyMPI_HAVE_MPI_Cart_get
#undef  PyMPI_HAVE_MPI_Cart_rank
#undef  PyMPI_HAVE_MPI_Cart_coords
#undef  PyMPI_HAVE_MPI_Cart_shift
#undef  PyMPI_HAVE_MPI_Cart_sub
#undef  PyMPI_HAVE_MPI_Cart_map
#undef  PyMPI_HAVE_MPI_Dims_create
#undef  PyMPI_HAVE_MPI_GRAPH
#undef  PyMPI_HAVE_MPI_Graph_create
#undef  PyMPI_HAVE_MPI_Graphdims_get
#undef  PyMPI_HAVE_MPI_Graph_get
#undef  PyMPI_HAVE_MPI_Graph_map
#undef  PyMPI_HAVE_MPI_Graph_neighbors_count
#undef  PyMPI_HAVE_MPI_Graph_neighbors
#undef  PyMPI_HAVE_MPI_DIST_GRAPH
#undef  PyMPI_HAVE_MPI_UNWEIGHTED
#undef  PyMPI_HAVE_MPI_WEIGHTS_EMPTY
#undef  PyMPI_HAVE_MPI_Dist_graph_create_adjacent
#undef  PyMPI_HAVE_MPI_Dist_graph_create
#undef  PyMPI_HAVE_MPI_Dist_graph_neighbors_count
#undef  PyMPI_HAVE_MPI_Dist_graph_neighbors
#undef  PyMPI_HAVE_MPI_Intercomm_create
#undef  PyMPI_HAVE_MPI_Intercomm_create_from_groups
#undef  PyMPI_HAVE_MPI_Comm_remote_group
#undef  PyMPI_HAVE_MPI_Comm_remote_size
#undef  PyMPI_HAVE_MPI_Intercomm_merge
#undef  PyMPI_HAVE_MPI_MAX_PORT_NAME
#undef  PyMPI_HAVE_MPI_Open_port
#undef  PyMPI_HAVE_MPI_Close_port
#undef  PyMPI_HAVE_MPI_Publish_name
#undef  PyMPI_HAVE_MPI_Unpublish_name
#undef  PyMPI_HAVE_MPI_Lookup_name
#undef  PyMPI_HAVE_MPI_Comm_accept
#undef  PyMPI_HAVE_MPI_Comm_connect
#undef  PyMPI_HAVE_MPI_Comm_join
#undef  PyMPI_HAVE_MPI_Comm_disconnect
#undef  PyMPI_HAVE_MPI_ARGV_NULL
#undef  PyMPI_HAVE_MPI_ARGVS_NULL
#undef  PyMPI_HAVE_MPI_ERRCODES_IGNORE
#undef  PyMPI_HAVE_MPI_Comm_spawn
#undef  PyMPI_HAVE_MPI_Comm_spawn_multiple
#undef  PyMPI_HAVE_MPI_Comm_get_parent
#undef  PyMPI_HAVE_MPI_Comm_get_name
#undef  PyMPI_HAVE_MPI_Comm_set_name
#undef  PyMPI_HAVE_MPI_TAG_UB
#undef  PyMPI_HAVE_MPI_IO
#undef  PyMPI_HAVE_MPI_WTIME_IS_GLOBAL
#undef  PyMPI_HAVE_MPI_UNIVERSE_SIZE
#undef  PyMPI_HAVE_MPI_APPNUM
#undef  PyMPI_HAVE_MPI_LASTUSEDCODE
#undef  PyMPI_HAVE_MPI_Comm_get_attr
#undef  PyMPI_HAVE_MPI_Comm_set_attr
#undef  PyMPI_HAVE_MPI_Comm_delete_attr
#undef  PyMPI_HAVE_MPI_Comm_copy_attr_function
#undef  PyMPI_HAVE_MPI_Comm_delete_attr_function
#undef  PyMPI_HAVE_MPI_COMM_DUP_FN
#undef  PyMPI_HAVE_MPI_COMM_NULL_COPY_FN
#undef  PyMPI_HAVE_MPI_COMM_NULL_DELETE_FN
#undef  PyMPI_HAVE_MPI_Comm_create_keyval
#undef  PyMPI_HAVE_MPI_Comm_free_keyval
#undef  PyMPI_HAVE_MPI_Comm_errhandler_fn
#undef  PyMPI_HAVE_MPI_Comm_errhandler_function
#undef  PyMPI_HAVE_MPI_Comm_create_errhandler
#undef  PyMPI_HAVE_MPI_Comm_get_errhandler
#undef  PyMPI_HAVE_MPI_Comm_set_errhandler
#undef  PyMPI_HAVE_MPI_Comm_call_errhandler
#undef  PyMPI_HAVE_MPI_Buffer_attach_c
#undef  PyMPI_HAVE_MPI_Buffer_detach_c
#undef  PyMPI_HAVE_MPI_Comm_attach_buffer_c
#undef  PyMPI_HAVE_MPI_Comm_detach_buffer_c
#undef  PyMPI_HAVE_MPI_Session_attach_buffer_c
#undef  PyMPI_HAVE_MPI_Session_detach_buffer_c
#undef  PyMPI_HAVE_MPI_Send_c
#undef  PyMPI_HAVE_MPI_Recv_c
#undef  PyMPI_HAVE_MPI_Sendrecv_c
#undef  PyMPI_HAVE_MPI_Sendrecv_replace_c
#undef  PyMPI_HAVE_MPI_Bsend_c
#undef  PyMPI_HAVE_MPI_Ssend_c
#undef  PyMPI_HAVE_MPI_Rsend_c
#undef  PyMPI_HAVE_MPI_Isend_c
#undef  PyMPI_HAVE_MPI_Irecv_c
#undef  PyMPI_HAVE_MPI_Isendrecv_c
#undef  PyMPI_HAVE_MPI_Isendrecv_replace_c
#undef  PyMPI_HAVE_MPI_Ibsend_c
#undef  PyMPI_HAVE_MPI_Issend_c
#undef  PyMPI_HAVE_MPI_Irsend_c
#undef  PyMPI_HAVE_MPI_Send_init_c
#undef  PyMPI_HAVE_MPI_Recv_init_c
#undef  PyMPI_HAVE_MPI_Bsend_init_c
#undef  PyMPI_HAVE_MPI_Ssend_init_c
#undef  PyMPI_HAVE_MPI_Rsend_init_c
#undef  PyMPI_HAVE_MPI_Mrecv_c
#undef  PyMPI_HAVE_MPI_Imrecv_c
#undef  PyMPI_HAVE_MPI_Bcast_c
#undef  PyMPI_HAVE_MPI_Gather_c
#undef  PyMPI_HAVE_MPI_Gatherv_c
#undef  PyMPI_HAVE_MPI_Scatter_c
#undef  PyMPI_HAVE_MPI_Scatterv_c
#undef  PyMPI_HAVE_MPI_Allgather_c
#undef  PyMPI_HAVE_MPI_Allgatherv_c
#undef  PyMPI_HAVE_MPI_Alltoall_c
#undef  PyMPI_HAVE_MPI_Alltoallv_c
#undef  PyMPI_HAVE_MPI_Alltoallw_c
#undef  PyMPI_HAVE_MPI_Reduce_local_c
#undef  PyMPI_HAVE_MPI_Reduce_c
#undef  PyMPI_HAVE_MPI_Allreduce_c
#undef  PyMPI_HAVE_MPI_Reduce_scatter_block_c
#undef  PyMPI_HAVE_MPI_Reduce_scatter_c
#undef  PyMPI_HAVE_MPI_Scan_c
#undef  PyMPI_HAVE_MPI_Exscan_c
#undef  PyMPI_HAVE_MPI_Neighbor_allgather_c
#undef  PyMPI_HAVE_MPI_Neighbor_allgatherv_c
#undef  PyMPI_HAVE_MPI_Neighbor_alltoall_c
#undef  PyMPI_HAVE_MPI_Neighbor_alltoallv_c
#undef  PyMPI_HAVE_MPI_Neighbor_alltoallw_c
#undef  PyMPI_HAVE_MPI_Ibcast_c
#undef  PyMPI_HAVE_MPI_Igather_c
#undef  PyMPI_HAVE_MPI_Igatherv_c
#undef  PyMPI_HAVE_MPI_Iscatter_c
#undef  PyMPI_HAVE_MPI_Iscatterv_c
#undef  PyMPI_HAVE_MPI_Iallgather_c
#undef  PyMPI_HAVE_MPI_Iallgatherv_c
#undef  PyMPI_HAVE_MPI_Ialltoall_c
#undef  PyMPI_HAVE_MPI_Ialltoallv_c
#undef  PyMPI_HAVE_MPI_Ialltoallw_c
#undef  PyMPI_HAVE_MPI_Ireduce_c
#undef  PyMPI_HAVE_MPI_Iallreduce_c
#undef  PyMPI_HAVE_MPI_Ireduce_scatter_block_c
#undef  PyMPI_HAVE_MPI_Ireduce_scatter_c
#undef  PyMPI_HAVE_MPI_Iscan_c
#undef  PyMPI_HAVE_MPI_Iexscan_c
#undef  PyMPI_HAVE_MPI_Ineighbor_allgather_c
#undef  PyMPI_HAVE_MPI_Ineighbor_allgatherv_c
#undef  PyMPI_HAVE_MPI_Ineighbor_alltoall_c
#undef  PyMPI_HAVE_MPI_Ineighbor_alltoallv_c
#undef  PyMPI_HAVE_MPI_Ineighbor_alltoallw_c
#undef  PyMPI_HAVE_MPI_Bcast_init_c
#undef  PyMPI_HAVE_MPI_Gather_init_c
#undef  PyMPI_HAVE_MPI_Gatherv_init_c
#undef  PyMPI_HAVE_MPI_Scatter_init_c
#undef  PyMPI_HAVE_MPI_Scatterv_init_c
#undef  PyMPI_HAVE_MPI_Allgather_init_c
#undef  PyMPI_HAVE_MPI_Allgatherv_init_c
#undef  PyMPI_HAVE_MPI_Alltoall_init_c
#undef  PyMPI_HAVE_MPI_Alltoallv_init_c
#undef  PyMPI_HAVE_MPI_Alltoallw_init_c
#undef  PyMPI_HAVE_MPI_Reduce_init_c
#undef  PyMPI_HAVE_MPI_Allreduce_init_c
#undef  PyMPI_HAVE_MPI_Reduce_scatter_block_init_c
#undef  PyMPI_HAVE_MPI_Reduce_scatter_init_c
#undef  PyMPI_HAVE_MPI_Scan_init_c
#undef  PyMPI_HAVE_MPI_Exscan_init_c
#undef  PyMPI_HAVE_MPI_Neighbor_allgather_init_c
#undef  PyMPI_HAVE_MPI_Neighbor_allgatherv_init_c
#undef  PyMPI_HAVE_MPI_Neighbor_alltoall_init_c
#undef  PyMPI_HAVE_MPI_Neighbor_alltoallv_init_c
#undef  PyMPI_HAVE_MPI_Neighbor_alltoallw_init_c
#undef  PyMPI_HAVE_MPI_WIN_NULL
#undef  PyMPI_HAVE_MPI_Win_free
#undef  PyMPI_HAVE_MPI_Win_create
#undef  PyMPI_HAVE_MPI_Win_allocate
#undef  PyMPI_HAVE_MPI_Win_allocate_shared
#undef  PyMPI_HAVE_MPI_Win_shared_query
#undef  PyMPI_HAVE_MPI_Win_create_dynamic
#undef  PyMPI_HAVE_MPI_Win_attach
#undef  PyMPI_HAVE_MPI_Win_detach
#undef  PyMPI_HAVE_MPI_Win_set_info
#undef  PyMPI_HAVE_MPI_Win_get_info
#undef  PyMPI_HAVE_MPI_Win_get_group
#undef  PyMPI_HAVE_MPI_Get
#undef  PyMPI_HAVE_MPI_Put
#undef  PyMPI_HAVE_MPI_Accumulate
#undef  PyMPI_HAVE_MPI_Get_accumulate
#undef  PyMPI_HAVE_MPI_Fetch_and_op
#undef  PyMPI_HAVE_MPI_Compare_and_swap
#undef  PyMPI_HAVE_MPI_Rget
#undef  PyMPI_HAVE_MPI_Rput
#undef  PyMPI_HAVE_MPI_Raccumulate
#undef  PyMPI_HAVE_MPI_Rget_accumulate
#undef  PyMPI_HAVE_MPI_MODE_NOCHECK
#undef  PyMPI_HAVE_MPI_MODE_NOSTORE
#undef  PyMPI_HAVE_MPI_MODE_NOPUT
#undef  PyMPI_HAVE_MPI_MODE_NOPRECEDE
#undef  PyMPI_HAVE_MPI_MODE_NOSUCCEED
#undef  PyMPI_HAVE_MPI_Win_fence
#undef  PyMPI_HAVE_MPI_Win_post
#undef  PyMPI_HAVE_MPI_Win_start
#undef  PyMPI_HAVE_MPI_Win_complete
#undef  PyMPI_HAVE_MPI_Win_wait
#undef  PyMPI_HAVE_MPI_Win_test
#undef  PyMPI_HAVE_MPI_LOCK_EXCLUSIVE
#undef  PyMPI_HAVE_MPI_LOCK_SHARED
#undef  PyMPI_HAVE_MPI_Win_lock
#undef  PyMPI_HAVE_MPI_Win_unlock
#undef  PyMPI_HAVE_MPI_Win_lock_all
#undef  PyMPI_HAVE_MPI_Win_unlock_all
#undef  PyMPI_HAVE_MPI_Win_flush
#undef  PyMPI_HAVE_MPI_Win_flush_all
#undef  PyMPI_HAVE_MPI_Win_flush_local
#undef  PyMPI_HAVE_MPI_Win_flush_local_all
#undef  PyMPI_HAVE_MPI_Win_sync
#undef  PyMPI_HAVE_MPI_Win_get_name
#undef  PyMPI_HAVE_MPI_Win_set_name
#undef  PyMPI_HAVE_MPI_WIN_BASE
#undef  PyMPI_HAVE_MPI_WIN_SIZE
#undef  PyMPI_HAVE_MPI_WIN_DISP_UNIT
#undef  PyMPI_HAVE_MPI_WIN_CREATE_FLAVOR
#undef  PyMPI_HAVE_MPI_WIN_MODEL
#undef  PyMPI_HAVE_MPI_WIN_FLAVOR_CREATE
#undef  PyMPI_HAVE_MPI_WIN_FLAVOR_ALLOCATE
#undef  PyMPI_HAVE_MPI_WIN_FLAVOR_DYNAMIC
#undef  PyMPI_HAVE_MPI_WIN_FLAVOR_SHARED
#undef  PyMPI_HAVE_MPI_WIN_SEPARATE
#undef  PyMPI_HAVE_MPI_WIN_UNIFIED
#undef  PyMPI_HAVE_MPI_Win_get_attr
#undef  PyMPI_HAVE_MPI_Win_set_attr
#undef  PyMPI_HAVE_MPI_Win_delete_attr
#undef  PyMPI_HAVE_MPI_Win_copy_attr_function
#undef  PyMPI_HAVE_MPI_Win_delete_attr_function
#undef  PyMPI_HAVE_MPI_WIN_DUP_FN
#undef  PyMPI_HAVE_MPI_WIN_NULL_COPY_FN
#undef  PyMPI_HAVE_MPI_WIN_NULL_DELETE_FN
#undef  PyMPI_HAVE_MPI_Win_create_keyval
#undef  PyMPI_HAVE_MPI_Win_free_keyval
#undef  PyMPI_HAVE_MPI_Win_errhandler_fn
#undef  PyMPI_HAVE_MPI_Win_errhandler_function
#undef  PyMPI_HAVE_MPI_Win_create_errhandler
#undef  PyMPI_HAVE_MPI_Win_get_errhandler
#undef  PyMPI_HAVE_MPI_Win_set_errhandler
#undef  PyMPI_HAVE_MPI_Win_call_errhandler
#undef  PyMPI_HAVE_MPI_Win_create_c
#undef  PyMPI_HAVE_MPI_Win_allocate_c
#undef  PyMPI_HAVE_MPI_Win_allocate_shared_c
#undef  PyMPI_HAVE_MPI_Win_shared_query_c
#undef  PyMPI_HAVE_MPI_Get_c
#undef  PyMPI_HAVE_MPI_Put_c
#undef  PyMPI_HAVE_MPI_Accumulate_c
#undef  PyMPI_HAVE_MPI_Get_accumulate_c
#undef  PyMPI_HAVE_MPI_Rget_c
#undef  PyMPI_HAVE_MPI_Rput_c
#undef  PyMPI_HAVE_MPI_Raccumulate_c
#undef  PyMPI_HAVE_MPI_Rget_accumulate_c
#undef  PyMPI_HAVE_MPI_FILE_NULL
#undef  PyMPI_HAVE_MPI_MODE_RDONLY
#undef  PyMPI_HAVE_MPI_MODE_RDWR
#undef  PyMPI_HAVE_MPI_MODE_WRONLY
#undef  PyMPI_HAVE_MPI_MODE_CREATE
#undef  PyMPI_HAVE_MPI_MODE_EXCL
#undef  PyMPI_HAVE_MPI_MODE_DELETE_ON_CLOSE
#undef  PyMPI_HAVE_MPI_MODE_UNIQUE_OPEN
#undef  PyMPI_HAVE_MPI_MODE_APPEND
#undef  PyMPI_HAVE_MPI_MODE_SEQUENTIAL
#undef  PyMPI_HAVE_MPI_File_open
#undef  PyMPI_HAVE_MPI_File_close
#undef  PyMPI_HAVE_MPI_File_delete
#undef  PyMPI_HAVE_MPI_File_set_size
#undef  PyMPI_HAVE_MPI_File_preallocate
#undef  PyMPI_HAVE_MPI_File_get_size
#undef  PyMPI_HAVE_MPI_File_get_group
#undef  PyMPI_HAVE_MPI_File_get_amode
#undef  PyMPI_HAVE_MPI_File_set_info
#undef  PyMPI_HAVE_MPI_File_get_info
#undef  PyMPI_HAVE_MPI_File_get_view
#undef  PyMPI_HAVE_MPI_File_set_view
#undef  PyMPI_HAVE_MPI_File_read_at
#undef  PyMPI_HAVE_MPI_File_read_at_all
#undef  PyMPI_HAVE_MPI_File_write_at
#undef  PyMPI_HAVE_MPI_File_write_at_all
#undef  PyMPI_HAVE_MPI_File_iread_at
#undef  PyMPI_HAVE_MPI_File_iread_at_all
#undef  PyMPI_HAVE_MPI_File_iwrite_at
#undef  PyMPI_HAVE_MPI_File_iwrite_at_all
#undef  PyMPI_HAVE_MPI_SEEK_SET
#undef  PyMPI_HAVE_MPI_SEEK_CUR
#undef  PyMPI_HAVE_MPI_SEEK_END
#undef  PyMPI_HAVE_MPI_DISPLACEMENT_CURRENT
#undef  PyMPI_HAVE_MPI_File_seek
#undef  PyMPI_HAVE_MPI_File_get_position
#undef  PyMPI_HAVE_MPI_File_get_byte_offset
#undef  PyMPI_HAVE_MPI_File_read
#undef  PyMPI_HAVE_MPI_File_read_all
#undef  PyMPI_HAVE_MPI_File_write
#undef  PyMPI_HAVE_MPI_File_write_all
#undef  PyMPI_HAVE_MPI_File_iread
#undef  PyMPI_HAVE_MPI_File_iread_all
#undef  PyMPI_HAVE_MPI_File_iwrite
#undef  PyMPI_HAVE_MPI_File_iwrite_all
#undef  PyMPI_HAVE_MPI_File_read_shared
#undef  PyMPI_HAVE_MPI_File_write_shared
#undef  PyMPI_HAVE_MPI_File_iread_shared
#undef  PyMPI_HAVE_MPI_File_iwrite_shared
#undef  PyMPI_HAVE_MPI_File_read_ordered
#undef  PyMPI_HAVE_MPI_File_write_ordered
#undef  PyMPI_HAVE_MPI_File_seek_shared
#undef  PyMPI_HAVE_MPI_File_get_position_shared
#undef  PyMPI_HAVE_MPI_File_read_at_all_begin
#undef  PyMPI_HAVE_MPI_File_read_at_all_end
#undef  PyMPI_HAVE_MPI_File_write_at_all_begin
#undef  PyMPI_HAVE_MPI_File_write_at_all_end
#undef  PyMPI_HAVE_MPI_File_read_all_begin
#undef  PyMPI_HAVE_MPI_File_read_all_end
#undef  PyMPI_HAVE_MPI_File_write_all_begin
#undef  PyMPI_HAVE_MPI_File_write_all_end
#undef  PyMPI_HAVE_MPI_File_read_ordered_begin
#undef  PyMPI_HAVE_MPI_File_read_ordered_end
#undef  PyMPI_HAVE_MPI_File_write_ordered_begin
#undef  PyMPI_HAVE_MPI_File_write_ordered_end
#undef  PyMPI_HAVE_MPI_File_get_type_extent
#undef  PyMPI_HAVE_MPI_File_set_atomicity
#undef  PyMPI_HAVE_MPI_File_get_atomicity
#undef  PyMPI_HAVE_MPI_File_sync
#undef  PyMPI_HAVE_MPI_File_errhandler_fn
#undef  PyMPI_HAVE_MPI_File_errhandler_function
#undef  PyMPI_HAVE_MPI_File_create_errhandler
#undef  PyMPI_HAVE_MPI_File_get_errhandler
#undef  PyMPI_HAVE_MPI_File_set_errhandler
#undef  PyMPI_HAVE_MPI_File_call_errhandler
#undef  PyMPI_HAVE_MPI_Datarep_conversion_function
#undef  PyMPI_HAVE_MPI_Datarep_extent_function
#undef  PyMPI_HAVE_MPI_CONVERSION_FN_NULL
#undef  PyMPI_HAVE_MPI_MAX_DATAREP_STRING
#undef  PyMPI_HAVE_MPI_Register_datarep
#undef  PyMPI_HAVE_MPI_File_read_at_c
#undef  PyMPI_HAVE_MPI_File_read_at_all_c
#undef  PyMPI_HAVE_MPI_File_write_at_c
#undef  PyMPI_HAVE_MPI_File_write_at_all_c
#undef  PyMPI_HAVE_MPI_File_iread_at_c
#undef  PyMPI_HAVE_MPI_File_iread_at_all_c
#undef  PyMPI_HAVE_MPI_File_iwrite_at_c
#undef  PyMPI_HAVE_MPI_File_iwrite_at_all_c
#undef  PyMPI_HAVE_MPI_File_read_c
#undef  PyMPI_HAVE_MPI_File_read_all_c
#undef  PyMPI_HAVE_MPI_File_write_c
#undef  PyMPI_HAVE_MPI_File_write_all_c
#undef  PyMPI_HAVE_MPI_File_iread_c
#undef  PyMPI_HAVE_MPI_File_iread_all_c
#undef  PyMPI_HAVE_MPI_File_iwrite_c
#undef  PyMPI_HAVE_MPI_File_iwrite_all_c
#undef  PyMPI_HAVE_MPI_File_read_shared_c
#undef  PyMPI_HAVE_MPI_File_write_shared_c
#undef  PyMPI_HAVE_MPI_File_iread_shared_c
#undef  PyMPI_HAVE_MPI_File_iwrite_shared_c
#undef  PyMPI_HAVE_MPI_File_read_ordered_c
#undef  PyMPI_HAVE_MPI_File_write_ordered_c
#undef  PyMPI_HAVE_MPI_File_read_at_all_begin_c
#undef  PyMPI_HAVE_MPI_File_write_at_all_begin_c
#undef  PyMPI_HAVE_MPI_File_read_all_begin_c
#undef  PyMPI_HAVE_MPI_File_write_all_begin_c
#undef  PyMPI_HAVE_MPI_File_read_ordered_begin_c
#undef  PyMPI_HAVE_MPI_File_write_ordered_begin_c
#undef  PyMPI_HAVE_MPI_File_get_type_extent_c
#undef  PyMPI_HAVE_MPI_Datarep_conversion_function_c
#undef  PyMPI_HAVE_MPI_CONVERSION_FN_NULL_C
#undef  PyMPI_HAVE_MPI_Register_datarep_c
#undef  PyMPI_HAVE_MPI_MAX_ERROR_STRING
#undef  PyMPI_HAVE_MPI_Error_class
#undef  PyMPI_HAVE_MPI_Error_string
#undef  PyMPI_HAVE_MPI_Add_error_class
#undef  PyMPI_HAVE_MPI_Remove_error_class
#undef  PyMPI_HAVE_MPI_Add_error_code
#undef  PyMPI_HAVE_MPI_Remove_error_code
#undef  PyMPI_HAVE_MPI_Add_error_string
#undef  PyMPI_HAVE_MPI_Remove_error_string
#undef  PyMPI_HAVE_MPI_SUCCESS
#undef  PyMPI_HAVE_MPI_ERR_LASTCODE
#undef  PyMPI_HAVE_MPI_ERR_ABI
#undef  PyMPI_HAVE_MPI_ERR_TYPE
#undef  PyMPI_HAVE_MPI_ERR_REQUEST
#undef  PyMPI_HAVE_MPI_ERR_OP
#undef  PyMPI_HAVE_MPI_ERR_GROUP
#undef  PyMPI_HAVE_MPI_ERR_INFO
#undef  PyMPI_HAVE_MPI_ERR_ERRHANDLER
#undef  PyMPI_HAVE_MPI_ERR_SESSION
#undef  PyMPI_HAVE_MPI_ERR_COMM
#undef  PyMPI_HAVE_MPI_ERR_WIN
#undef  PyMPI_HAVE_MPI_ERR_FILE
#undef  PyMPI_HAVE_MPI_ERR_BUFFER
#undef  PyMPI_HAVE_MPI_ERR_COUNT
#undef  PyMPI_HAVE_MPI_ERR_TAG
#undef  PyMPI_HAVE_MPI_ERR_RANK
#undef  PyMPI_HAVE_MPI_ERR_ROOT
#undef  PyMPI_HAVE_MPI_ERR_TRUNCATE
#undef  PyMPI_HAVE_MPI_ERR_IN_STATUS
#undef  PyMPI_HAVE_MPI_ERR_PENDING
#undef  PyMPI_HAVE_MPI_ERR_TOPOLOGY
#undef  PyMPI_HAVE_MPI_ERR_DIMS
#undef  PyMPI_HAVE_MPI_ERR_ARG
#undef  PyMPI_HAVE_MPI_ERR_OTHER
#undef  PyMPI_HAVE_MPI_ERR_UNKNOWN
#undef  PyMPI_HAVE_MPI_ERR_INTERN
#undef  PyMPI_HAVE_MPI_ERR_KEYVAL
#undef  PyMPI_HAVE_MPI_ERR_NO_MEM
#undef  PyMPI_HAVE_MPI_ERR_INFO_KEY
#undef  PyMPI_HAVE_MPI_ERR_INFO_VALUE
#undef  PyMPI_HAVE_MPI_ERR_INFO_NOKEY
#undef  PyMPI_HAVE_MPI_ERR_SPAWN
#undef  PyMPI_HAVE_MPI_ERR_PORT
#undef  PyMPI_HAVE_MPI_ERR_SERVICE
#undef  PyMPI_HAVE_MPI_ERR_NAME
#undef  PyMPI_HAVE_MPI_ERR_PROC_ABORTED
#undef  PyMPI_HAVE_MPI_ERR_BASE
#undef  PyMPI_HAVE_MPI_ERR_SIZE
#undef  PyMPI_HAVE_MPI_ERR_DISP
#undef  PyMPI_HAVE_MPI_ERR_ASSERT
#undef  PyMPI_HAVE_MPI_ERR_LOCKTYPE
#undef  PyMPI_HAVE_MPI_ERR_RMA_CONFLICT
#undef  PyMPI_HAVE_MPI_ERR_RMA_SYNC
#undef  PyMPI_HAVE_MPI_ERR_RMA_RANGE
#undef  PyMPI_HAVE_MPI_ERR_RMA_ATTACH
#undef  PyMPI_HAVE_MPI_ERR_RMA_SHARED
#undef  PyMPI_HAVE_MPI_ERR_RMA_FLAVOR
#undef  PyMPI_HAVE_MPI_ERR_BAD_FILE
#undef  PyMPI_HAVE_MPI_ERR_NO_SUCH_FILE
#undef  PyMPI_HAVE_MPI_ERR_FILE_EXISTS
#undef  PyMPI_HAVE_MPI_ERR_FILE_IN_USE
#undef  PyMPI_HAVE_MPI_ERR_AMODE
#undef  PyMPI_HAVE_MPI_ERR_ACCESS
#undef  PyMPI_HAVE_MPI_ERR_READ_ONLY
#undef  PyMPI_HAVE_MPI_ERR_NO_SPACE
#undef  PyMPI_HAVE_MPI_ERR_QUOTA
#undef  PyMPI_HAVE_MPI_ERR_UNSUPPORTED_OPERATION
#undef  PyMPI_HAVE_MPI_ERR_NOT_SAME
#undef  PyMPI_HAVE_MPI_ERR_IO
#undef  PyMPI_HAVE_MPI_ERR_UNSUPPORTED_DATAREP
#undef  PyMPI_HAVE_MPI_ERR_CONVERSION
#undef  PyMPI_HAVE_MPI_ERR_DUP_DATAREP
#undef  PyMPI_HAVE_MPI_ERR_VALUE_TOO_LARGE
#undef  PyMPI_HAVE_MPI_Alloc_mem
#undef  PyMPI_HAVE_MPI_Free_mem
#define PyMPI_HAVE_MPI_Init 1
#define PyMPI_HAVE_MPI_Finalize 1
#define PyMPI_HAVE_MPI_Initialized 1
#define PyMPI_HAVE_MPI_Finalized 1
#undef  PyMPI_HAVE_MPI_THREAD_SINGLE
#undef  PyMPI_HAVE_MPI_THREAD_FUNNELED
#undef  PyMPI_HAVE_MPI_THREAD_SERIALIZED
#undef  PyMPI_HAVE_MPI_THREAD_MULTIPLE
#undef  PyMPI_HAVE_MPI_Init_thread
#undef  PyMPI_HAVE_MPI_Query_thread
#undef  PyMPI_HAVE_MPI_Is_thread_main
#undef  PyMPI_HAVE_MPI_VERSION
#undef  PyMPI_HAVE_MPI_SUBVERSION
#undef  PyMPI_HAVE_MPI_Get_version
#undef  PyMPI_HAVE_MPI_MAX_LIBRARY_VERSION_STRING
#undef  PyMPI_HAVE_MPI_Get_library_version
#undef  PyMPI_HAVE_MPI_Abi_get_version
#undef  PyMPI_HAVE_MPI_Abi_get_info
#undef  PyMPI_HAVE_MPI_Abi_get_fortran_info
#undef  PyMPI_HAVE_MPI_MAX_PROCESSOR_NAME
#undef  PyMPI_HAVE_MPI_Get_processor_name
#undef  PyMPI_HAVE_MPI_Get_hw_resource_info
#undef  PyMPI_HAVE_MPI_Wtime
#undef  PyMPI_HAVE_MPI_Wtick
#undef  PyMPI_HAVE_MPI_Pcontrol
#undef  PyMPI_HAVE_MPI_Fint
#undef  PyMPI_HAVE_MPI_F_SOURCE
#undef  PyMPI_HAVE_MPI_F_TAG
#undef  PyMPI_HAVE_MPI_F_ERROR
#undef  PyMPI_HAVE_MPI_F_STATUS_SIZE
#undef  PyMPI_HAVE_MPI_F_STATUS_IGNORE
#undef  PyMPI_HAVE_MPI_F_STATUSES_IGNORE
#undef  PyMPI_HAVE_MPI_Status_c2f
#undef  PyMPI_HAVE_MPI_Status_f2c
#undef  PyMPI_HAVE_MPI_Type_c2f
#undef  PyMPI_HAVE_MPI_Request_c2f
#undef  PyMPI_HAVE_MPI_Message_c2f
#undef  PyMPI_HAVE_MPI_Op_c2f
#undef  PyMPI_HAVE_MPI_Group_c2f
#undef  PyMPI_HAVE_MPI_Info_c2f
#undef  PyMPI_HAVE_MPI_Session_c2f
#undef  PyMPI_HAVE_MPI_Comm_c2f
#undef  PyMPI_HAVE_MPI_Win_c2f
#undef  PyMPI_HAVE_MPI_File_c2f
#undef  PyMPI_HAVE_MPI_Errhandler_c2f
#undef  PyMPI_HAVE_MPI_Type_f2c
#undef  PyMPI_HAVE_MPI_Request_f2c
#undef  PyMPI_HAVE_MPI_Message_f2c
#undef  PyMPI_HAVE_MPI_Op_f2c
#undef  PyMPI_HAVE_MPI_Group_f2c
#undef  PyMPI_HAVE_MPI_Info_f2c
#undef  PyMPI_HAVE_MPI_Session_f2c
#undef  PyMPI_HAVE_MPI_Comm_f2c
#undef  PyMPI_HAVE_MPI_Win_f2c
#undef  PyMPI_HAVE_MPI_File_f2c
#undef  PyMPI_HAVE_MPI_Errhandler_f2c
#undef  PyMPI_HAVE_MPI_Comm_toint
#undef  PyMPI_HAVE_MPI_Errhandler_toint
#undef  PyMPI_HAVE_MPI_File_toint
#undef  PyMPI_HAVE_MPI_Group_toint
#undef  PyMPI_HAVE_MPI_Info_toint
#undef  PyMPI_HAVE_MPI_Message_toint
#undef  PyMPI_HAVE_MPI_Op_toint
#undef  PyMPI_HAVE_MPI_Request_toint
#undef  PyMPI_HAVE_MPI_Session_toint
#undef  PyMPI_HAVE_MPI_Type_toint
#undef  PyMPI_HAVE_MPI_Win_toint
#undef  PyMPI_HAVE_MPI_Comm_fromint
#undef  PyMPI_HAVE_MPI_Errhandler_fromint
#undef  PyMPI_HAVE_MPI_File_fromint
#undef  PyMPI_HAVE_MPI_Group_fromint
#undef  PyMPI_HAVE_MPI_Info_fromint
#undef  PyMPI_HAVE_MPI_Message_fromint
#undef  PyMPI_HAVE_MPI_Op_fromint
#undef  PyMPI_HAVE_MPI_Request_fromint
#undef  PyMPI_HAVE_MPI_Session_fromint
#undef  PyMPI_HAVE_MPI_Type_fromint
#undef  PyMPI_HAVE_MPI_Win_fromint
#undef  PyMPI_HAVE_MPI_HOST
#undef  PyMPI_HAVE_MPI_Info_get
#undef  PyMPI_HAVE_MPI_Info_get_valuelen
#undef  PyMPI_HAVE_MPI_Attr_get
#undef  PyMPI_HAVE_MPI_Attr_put
#undef  PyMPI_HAVE_MPI_Attr_delete
#undef  PyMPI_HAVE_MPI_Copy_function
#undef  PyMPI_HAVE_MPI_Delete_function
#undef  PyMPI_HAVE_MPI_DUP_FN
#undef  PyMPI_HAVE_MPI_NULL_COPY_FN
#undef  PyMPI_HAVE_MPI_NULL_DELETE_FN
#undef  PyMPI_HAVE_MPI_Keyval_create
#undef  PyMPI_HAVE_MPI_Keyval_free
#undef  PyMPI_HAVE_MPI_Errhandler_get
#undef  PyMPI_HAVE_MPI_Errhandler_set
#undef  PyMPI_HAVE_MPI_Handler_function
#undef  PyMPI_HAVE_MPI_Errhandler_create
#undef  PyMPI_HAVE_MPI_Address
#undef  PyMPI_HAVE_MPI_UB
#undef  PyMPI_HAVE_MPI_LB
#undef  PyMPI_HAVE_MPI_Type_lb
#undef  PyMPI_HAVE_MPI_Type_ub
#undef  PyMPI_HAVE_MPI_Type_extent
#undef  PyMPI_HAVE_MPI_Type_hvector
#undef  PyMPI_HAVE_MPI_Type_hindexed
#undef  PyMPI_HAVE_MPI_Type_struct
#undef  PyMPI_HAVE_MPI_COMBINER_HVECTOR_INTEGER
#undef  PyMPI_HAVE_MPI_COMBINER_HINDEXED_INTEGER
#undef  PyMPI_HAVE_MPI_COMBINER_STRUCT_INTEGER
#undef  PyMPI_HAVE_MPI_ERR_REVOKED
#undef  PyMPI_HAVE_MPI_ERR_PROC_FAILED
#undef  PyMPI_HAVE_MPI_ERR_PROC_FAILED_PENDING
#undef  PyMPI_HAVE_MPI_Comm_revoke
#undef  PyMPI_HAVE_MPI_Comm_is_revoked
#undef  PyMPI_HAVE_MPI_Comm_get_failed
#undef  PyMPI_HAVE_MPI_Comm_ack_failed
#undef  PyMPI_HAVE_MPI_Comm_agree
#undef  PyMPI_HAVE_MPI_Comm_iagree
#undef  PyMPI_HAVE_MPI_Comm_shrink
#undef  PyMPI_HAVE_MPI_Comm_ishrink

#endif /* PyMPI_PYMPICONF_H */
