/* Author:  Lisandro Dalcin   */
/* Contact: dalcinl@gmail.com */
#ifndef PyMPI_LIBMPI_H
#define PyMPI_LIBMPI_H

#include "lib-mpi/config.h"
#include "lib-mpi/defines.h"
#include "lib-mpi/missing.h"
#include "lib-mpi/fallback.h"
#include "lib-mpi/mpiabi.h"
#include "lib-mpi/mpiapi.h"
#include "lib-mpi/compat.h"

#endif /* PyMPI_LIBMPI_H */
