#ifndef PyMPI_CONFIG_MPIABI_H
#define PyMPI_CONFIG_MPIABI_H

#include "mpiapi.h"

#undef PyMPI_HAVE_MPI_Fint
#undef PyMPI_HAVE_MPI_Comm_c2f
#undef PyMPI_HAVE_MPI_Errhandler_c2f
#undef PyMPI_HAVE_MPI_File_c2f
#undef PyMPI_HAVE_MPI_Group_c2f
#undef PyMPI_HAVE_MPI_Info_c2f
#undef PyMPI_HAVE_MPI_Message_c2f
#undef PyMPI_HAVE_MPI_Op_c2f
#undef PyMPI_HAVE_MPI_Request_c2f
#undef PyMPI_HAVE_MPI_Status_c2f
#undef PyMPI_HAVE_MPI_Session_c2f
#undef PyMPI_HAVE_MPI_Type_c2f
#undef PyMPI_HAVE_MPI_Win_c2f
#undef PyMPI_HAVE_MPI_Comm_f2c
#undef PyMPI_HAVE_MPI_Errhandler_f2c
#undef PyMPI_HAVE_MPI_File_f2c
#undef PyMPI_HAVE_MPI_Group_f2c
#undef PyMPI_HAVE_MPI_Info_f2c
#undef PyMPI_HAVE_MPI_Message_f2c
#undef PyMPI_HAVE_MPI_Op_f2c
#undef PyMPI_HAVE_MPI_Request_f2c
#undef PyMPI_HAVE_MPI_Status_f2c
#undef PyMPI_HAVE_MPI_Session_f2c
#undef PyMPI_HAVE_MPI_Type_f2c
#undef PyMPI_HAVE_MPI_Win_f2c

#endif /* !PyMPI_CONFIG_MPIABI_H */
