#undef PyMPI_HAVE_MPI_File
#undef PyMPI_HAVE_MPI_FILE_NULL
#undef PyMPI_HAVE_MPI_MODE_RDONLY
#undef PyMPI_HAVE_MPI_MODE_RDWR
#undef PyMPI_HAVE_MPI_MODE_WRONLY
#undef PyMPI_HAVE_MPI_MODE_CREATE
#undef PyMPI_HAVE_MPI_MODE_EXCL
#undef PyMPI_HAVE_MPI_MODE_DELETE_ON_CLOSE
#undef PyMPI_HAVE_MPI_MODE_UNIQUE_OPEN
#undef PyMPI_HAVE_MPI_MODE_APPEND
#undef PyMPI_HAVE_MPI_MODE_SEQUENTIAL
#undef PyMPI_HAVE_MPI_File_open
#undef PyMPI_HAVE_MPI_File_close
#undef PyMPI_HAVE_MPI_File_delete
#undef PyMPI_HAVE_MPI_File_set_size
#undef PyMPI_HAVE_MPI_File_preallocate
#undef PyMPI_HAVE_MPI_File_get_size
#undef PyMPI_HAVE_MPI_File_get_group
#undef PyMPI_HAVE_MPI_File_get_amode
#undef PyMPI_HAVE_MPI_File_set_info
#undef PyMPI_HAVE_MPI_File_get_info
#undef PyMPI_HAVE_MPI_File_get_view
#undef PyMPI_HAVE_MPI_File_set_view
#undef PyMPI_HAVE_MPI_File_read_at
#undef PyMPI_HAVE_MPI_File_read_at_all
#undef PyMPI_HAVE_MPI_File_write_at
#undef PyMPI_HAVE_MPI_File_write_at_all
#undef PyMPI_HAVE_MPI_File_iread_at
#undef PyMPI_HAVE_MPI_File_iread_at_all
#undef PyMPI_HAVE_MPI_File_iwrite_at
#undef PyMPI_HAVE_MPI_File_iwrite_at_all
#undef PyMPI_HAVE_MPI_SEEK_SET
#undef PyMPI_HAVE_MPI_SEEK_CUR
#undef PyMPI_HAVE_MPI_SEEK_END
#undef PyMPI_HAVE_MPI_DISPLACEMENT_CURRENT
#undef PyMPI_HAVE_MPI_File_seek
#undef PyMPI_HAVE_MPI_File_get_position
#undef PyMPI_HAVE_MPI_File_get_byte_offset
#undef PyMPI_HAVE_MPI_File_read
#undef PyMPI_HAVE_MPI_File_read_all
#undef PyMPI_HAVE_MPI_File_write
#undef PyMPI_HAVE_MPI_File_write_all
#undef PyMPI_HAVE_MPI_File_iread
#undef PyMPI_HAVE_MPI_File_iread_all
#undef PyMPI_HAVE_MPI_File_iwrite
#undef PyMPI_HAVE_MPI_File_iwrite_all
#undef PyMPI_HAVE_MPI_File_read_shared
#undef PyMPI_HAVE_MPI_File_write_shared
#undef PyMPI_HAVE_MPI_File_iread_shared
#undef PyMPI_HAVE_MPI_File_iwrite_shared
#undef PyMPI_HAVE_MPI_File_read_ordered
#undef PyMPI_HAVE_MPI_File_write_ordered
#undef PyMPI_HAVE_MPI_File_seek_shared
#undef PyMPI_HAVE_MPI_File_get_position_shared
#undef PyMPI_HAVE_MPI_File_read_at_all_begin
#undef PyMPI_HAVE_MPI_File_read_at_all_end
#undef PyMPI_HAVE_MPI_File_write_at_all_begin
#undef PyMPI_HAVE_MPI_File_write_at_all_end
#undef PyMPI_HAVE_MPI_File_read_all_begin
#undef PyMPI_HAVE_MPI_File_read_all_end
#undef PyMPI_HAVE_MPI_File_write_all_begin
#undef PyMPI_HAVE_MPI_File_write_all_end
#undef PyMPI_HAVE_MPI_File_read_ordered_begin
#undef PyMPI_HAVE_MPI_File_read_ordered_end
#undef PyMPI_HAVE_MPI_File_write_ordered_begin
#undef PyMPI_HAVE_MPI_File_write_ordered_end
#undef PyMPI_HAVE_MPI_File_get_type_extent
#undef PyMPI_HAVE_MPI_File_set_atomicity
#undef PyMPI_HAVE_MPI_File_get_atomicity
#undef PyMPI_HAVE_MPI_File_sync
#undef PyMPI_HAVE_MPI_File_get_errhandler
#undef PyMPI_HAVE_MPI_File_set_errhandler
#undef PyMPI_HAVE_MPI_File_errhandler_fn
#undef PyMPI_HAVE_MPI_File_errhandler_function
#undef PyMPI_HAVE_MPI_File_create_errhandler
#undef PyMPI_HAVE_MPI_File_call_errhandler

#undef PyMPI_HAVE_MPI_MAX_DATAREP_STRING
#undef PyMPI_HAVE_MPI_Datarep_extent_function
#undef PyMPI_HAVE_MPI_Datarep_conversion_function
#undef PyMPI_HAVE_MPI_CONVERSION_FN_NULL
#undef PyMPI_HAVE_MPI_Register_datarep

#undef PyMPI_HAVE_MPI_File_read_at_c
#undef PyMPI_HAVE_MPI_File_read_at_all_c
#undef PyMPI_HAVE_MPI_File_write_at_c
#undef PyMPI_HAVE_MPI_File_write_at_all_c
#undef PyMPI_HAVE_MPI_File_iread_at_c
#undef PyMPI_HAVE_MPI_File_iread_at_all_c
#undef PyMPI_HAVE_MPI_File_iwrite_at_c
#undef PyMPI_HAVE_MPI_File_iwrite_at_all_c
#undef PyMPI_HAVE_MPI_File_read_c
#undef PyMPI_HAVE_MPI_File_read_all_c
#undef PyMPI_HAVE_MPI_File_write_c
#undef PyMPI_HAVE_MPI_File_write_all_c
#undef PyMPI_HAVE_MPI_File_iread_c
#undef PyMPI_HAVE_MPI_File_iread_all_c
#undef PyMPI_HAVE_MPI_File_iwrite_c
#undef PyMPI_HAVE_MPI_File_iwrite_all_c
#undef PyMPI_HAVE_MPI_File_read_shared_c
#undef PyMPI_HAVE_MPI_File_write_shared_c
#undef PyMPI_HAVE_MPI_File_iread_shared_c
#undef PyMPI_HAVE_MPI_File_iwrite_shared_c
#undef PyMPI_HAVE_MPI_File_read_ordered_c
#undef PyMPI_HAVE_MPI_File_write_ordered_c
#undef PyMPI_HAVE_MPI_File_read_at_all_begin_c
#undef PyMPI_HAVE_MPI_File_write_at_all_begin_c
#undef PyMPI_HAVE_MPI_File_read_all_begin_c
#undef PyMPI_HAVE_MPI_File_write_all_begin_c
#undef PyMPI_HAVE_MPI_File_read_ordered_begin_c
#undef PyMPI_HAVE_MPI_File_write_ordered_begin_c
#undef PyMPI_HAVE_MPI_File_get_type_extent_c

#undef PyMPI_HAVE_MPI_Datarep_conversion_function_c
#undef PyMPI_HAVE_MPI_CONVERSION_FN_NULL_C
#undef PyMPI_HAVE_MPI_Register_datarep_c

#undef PyMPI_HAVE_MPI_File_c2f
#undef PyMPI_HAVE_MPI_File_f2c

#if !defined(MPI_ERR_FILE)
#undef PyMPI_HAVE_MPI_ERR_FILE
#undef PyMPI_HAVE_MPI_ERR_NOT_SAME
#undef PyMPI_HAVE_MPI_ERR_BAD_FILE
#undef PyMPI_HAVE_MPI_ERR_NO_SUCH_FILE
#undef PyMPI_HAVE_MPI_ERR_FILE_EXISTS
#undef PyMPI_HAVE_MPI_ERR_FILE_IN_USE
#undef PyMPI_HAVE_MPI_ERR_AMODE
#undef PyMPI_HAVE_MPI_ERR_ACCESS
#undef PyMPI_HAVE_MPI_ERR_READ_ONLY
#undef PyMPI_HAVE_MPI_ERR_NO_SPACE
#undef PyMPI_HAVE_MPI_ERR_QUOTA
#undef PyMPI_HAVE_MPI_ERR_UNSUPPORTED_DATAREP
#undef PyMPI_HAVE_MPI_ERR_UNSUPPORTED_OPERATION
#undef PyMPI_HAVE_MPI_ERR_CONVERSION
#undef PyMPI_HAVE_MPI_ERR_DUP_DATAREP
#undef PyMPI_HAVE_MPI_ERR_IO
#endif
